/*
 * Copyright (c) 2024 DTB <trinity@trinity.moe>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

#include <errno.h> /* errno */
#include <signal.h> /* signal(2), SIG_ERR, SIG_IGN, SIGINT */
#include <stdio.h> /* fclose(3), fopen(3), fprintf(3), getc(3), putc(3),
                    * setvbuf(3), size_t, _IONBF, NULL */
#include <stdlib.h> /* free(3), malloc(3), realloc(3) */
#include <string.h> /* strerror(3) */
#include <unistd.h> /* getopt(3) */
#if !defined EX_OSERR || !defined EX_USAGE
#	include <sysexits.h>
#endif
extern int errno;

/* This structure is how open files are tracked. */
struct Files{
	size_t a;	/* allocation */
	size_t s;	/* used size */
	char *mode;	/* file opening mode */
	char **names;	/* file names */
	FILE **files;	/* file pointers */
};

/* How much to grow the allocation when it's saturated. */
#ifndef ALLOC_INCREMENT
#	define ALLOC_INCREMENT 1
#endif

/* How much to grow the allocation at program start. */
#ifndef ALLOC_INITIAL
#	define ALLOC_INITIAL 10
#endif

/* pre-allocated strings */
static char *program_name = "<no argv[0]>";
static char *stdin_name = "<stdin>";
static char *stdout_name = "<stdout>";
static char *stderr_name = "<stderr>";
static char *(fmode[]) = { (char []){"rb"}, (char []){"rb+"} };
static char *wharsh = "wb";

/* Opens the file at the path p and puts it in the files struct, returning NULL
 * if either the opening or the placement of the open FILE pointer fail. */
#define Files_open(files, p) \
	Files_append((files), fopen((p), (files)->mode), (p))

/* Adds the open FILE pointer for the file at the path s to the files struct,
 * returning the FILE if successful and NULL if not, allocating more memory in
 * the files buffers as needed. */
static FILE *
Files_append(struct Files *files, FILE *file, char *name){

	if(files->s == -1)
		files->s = 0;

	if(file == NULL || (files->s == files->a
			&& ((files->files = realloc(files->files,
					(files->a += (files->a == 0)
						? ALLOC_INITIAL
						: ALLOC_INCREMENT)
					* sizeof *(files->files))) == NULL
				|| (files->names = realloc(files->names,
					files->a * sizeof *(files->names)))
					== NULL)))
		return NULL;

	files->names[files->s] = name;
	return files->files[files->s++] = file;
}

/* Prints a diagnostic message based on errno and returns an exit status
 * appropriate for an OS error. */
static int
oserr(char *s, char *r){

	fprintf(stderr, "%s: %s: %s\n", s, r, strerror(errno));

	return EX_OSERR;
}

/* Destructs the files[2] struct used by main by closing its files and freeing
 * its files and names arrays. Returns the destructed files. */
static struct Files *
terminate(struct Files *files){
	size_t i;
	size_t j;

	for(i = 0; i < 2; ++i){
		for(j = 0; j < files[i].s; ++j)
				if(files[i].files[j] != stdin
						&& files[i].files[j] != stdout
						&& files[i].files[j] != stderr)
					fclose(files[i].files[j]);
			free(files[i].files);
			free(files[i].names);
	}

	return files;
}

/* Prints a diagnostic message synopsizing usage and returns an exit status
 * appropriate for a usage error. */
static int
usage(char *s){

	fprintf(stderr, "Usage: %s (-aehnu)"
		" (-i [input])... (-o [output])...\n", s);

	return EX_USAGE;
}

int main(int argc, char *argv[]){
	int c;
	size_t i;
	size_t j;
	size_t k;
	struct Files files[2]; /* {read, write} */
	int s; /* scratch variable */

	/* The simple invocation (without given arguments) requires no memory
	 * allocations or use of the io structs and is therefore treated as a
	 * special case. To do the same with all bells and whistles call
	 * $ mm -i - -o - */
	if(argc < 2){ /* simple invocation */
		while((c = getc(stdin)) != EOF)
			if(putc(c, stdout) == EOF)
				return 1;
		return 0;
	}

	/* Initializes the files structs with their default values, standard
	 * input and standard output. Both files->s are set to -1, a special
	 * value indicating that they're at their defaults and that their FILE
	 * pointers won't need to be closed. If an input or an output is
	 * specified these initial values will be overwritten, so to, say, use
	 * mm(1) equivalently to tee(1p), -o - will need to be specified before
	 * additional files to ensure standard output is still written. */
	for(i = 0; i < 2; ++i){
		files[i].a = 0;
		files[i].s = 0;
		files[i].mode = fmode[i];
		files[i].files = NULL;
		files[i].names = NULL;
		Files_append(&files[i], i == 0 ? stdin : stdout,
			i == 0 ? stdin_name : stdout_name);
		files[i].s = -1;
	}

	s = 0; /* Refers to whether or not files will be unbuffered. */

	while((c = getopt(argc, argv, "aehi:no:u")) != -1)
		switch(c){
		case 'a': /* "rb+" -> "ab" */
			files[1].mode[0] = 'a';
			files[1].mode[2] = '\0';
			break;
		case 'e':
			if(Files_append(&files[1], stderr, stderr_name)
					!= NULL)
				break;
			terminate(files);
			return oserr(argv[0], "-e");
		case 'i':
			if(optarg[0] == '-' && optarg[1] == '\0'){ /* "-" */
				if(Files_append(&files[0], stdin, stdin_name)
						!= NULL)
					break;
				else
					optarg = stdin_name;
			}else if(Files_open(&files[0], optarg) != NULL)
				break;
			terminate(files);
			return oserr(argv[0], optarg);
		case 'o':
			if(optarg[0] == '-' && optarg[1] == '\0'){
				/* "-" */
				if(Files_append(&files[1], stdout,
						stdout_name) != NULL)
					break;
				optarg = stdout_name;
			}else if(Files_open(&files[1], optarg) != NULL)
				break;
			/* does not exist, so try to create it */
			if(errno == ENOENT){
				files[1].mode = wharsh;
				if(Files_open(&files[1], optarg) != NULL){
					files[1].mode = fmode[1];
					break;
				}
			}
			terminate(files);
			return oserr(argv[0], optarg);;
		case 'n':
			if(signal(SIGINT, SIG_IGN) != SIG_ERR)
				break;
			terminate(files);
			return oserr(argv[0], "-n");
		case 'u':
			s = 1;
			break;
		default:
			terminate(files);
			return usage(argv[0]);
		}

	/* Unbuffer files. */
	if(s){	for(	i = 0;
			i < files[0].s;
			setvbuf(files[0].files[i++], NULL, _IONBF, 0));
		for(	i = 0;
			i < files[1].s;
			setvbuf(files[1].files[i++], NULL, _IONBF, 0));
	}

	if(files[0].s == -1)
		files[0].s = 1;
	if(files[1].s == -1)
		files[1].s = 1;

	/* Reflects whether (s==0) or not (s==1) the full length of the inputs
	 * could be written to all outputs, which is the return value for main
	 * to maintain feature parity with tee(1p). */
	s = 0;

	/* Actual program loop. */
	for(i = 0; i < files[0].s; ++i) /* iterate ins */
		while((c = getc(files[0].files[i])) != EOF) /* iterate chars */
			for(j = 0; j < files[1].s; ++j) /* iterate outs */
				/* notebook's full */
				if(putc(c, files[1].files[j]) == EOF){
					s = 1;
					if(fclose(files[1].files[j]) == EOF)
						fprintf(stderr, "%s: %s: %s\n",
							program_name,
							files[1].names[j],
							strerror(errno));
					/* massage out the tense muscle */
					for(k = j; k < files[1].s - 1; ++k){
						files[1].files[k]
							= files[1].files[k+1];
						files[1].names[k]
							= files[1].names[k+1];
					}
					if(--files[i].s == 0){
						terminate(files);
						return 1;
					}
				}

	terminate(files);

	return s;
}
